from dataclasses import dataclass

from .gaps import Endpoint, SupportsLessThan

__all__ = ["Var", "x"]


@dataclass
class Var[T: SupportsLessThan]:
    """A very convenient constructor for `Endpoints`.

    It's best to start with an example::
        >>> from mind_the_gaps import Gaps, Var
        >>> x = Var("x")
        >>> 0 <= x
        Endpoint(value=0, boundary='[')
        >>> x < 1
        Endpoint(value=1, boundary=')')
        >>> print(Gaps([0 <= x, x < 1]))
        {[0, 1)}

    Values compared with a Var will return an endpoint. The type of comparison determines
    the boundary of the endpoint. `>` is left-open, `>=` is left-closed, `<` is right-open,
    and `<=` is right-closed.

    `Var("x")` can be imported directly as `x`, e.g., `from mind_the_gaps import x`.
    """

    name: str

    def __str__(self):
        return self.name

    def __gt__(self, value: T) -> Endpoint:
        return Endpoint(value, "(")

    def __ge__(self, value: T) -> Endpoint:
        return Endpoint(value, "[")

    def __lt__(self, value: T) -> Endpoint:
        return Endpoint(value, ")")

    def __le__(self, value: T) -> Endpoint:
        return Endpoint(value, "]")


x = Var("x")
"""A very convenient constructor for `Endpoints`.

It's best to start with an example::
    >>> from mind_the_gaps import Gaps, x
    >>> 0 <= x
    Endpoint(value=0, boundary='[')
    >>> x < 1
    Endpoint(value=1, boundary=')')
    >>> print(Gaps([0 <= x, x < 1]))
    {[0, 1)}

Values compared with `x` will return an endpoint. The type of comparison determines
the boundary of the endpoint. `>` is left-open, `>=` is left-closed, `<` is right-open,
and `<=` is right-closed.
"""



from __future__ import annotations

from bisect import bisect
from collections.abc import Callable
from dataclasses import dataclass, field
from functools import total_ordering
from operator import and_, attrgetter, or_, xor
from typing import Any, Final, Literal, Protocol, Self

__all__ = ["Endpoint", "Gaps"]

Boundary = Literal["(", ")", "[", "]"]
"""The boundary of an endpoint.

`"("` is open and left, `")"` is open and right,`"["` is closed and left, and
`"]"` is closed and right.
"""

_BOUNDARY_ORDER: Final[dict[Boundary, int]] = {")": 0, "[": 1, "]": 1, "(": 2}


def sub(a: bool, b: bool) -> bool:
    """`a` and not `b`."""
    return a and not b


class SupportsLessThan(Protocol):
    """Supports the less-than (`<`) operator."""

    def __lt__(self, other: Any, /) -> bool: ...


@total_ordering
@dataclass(frozen=True, slots=True)
class Endpoint[T: SupportsLessThan]:
    """An interval endpoint."""

    value: T
    """Value of endpoint."""
    boundary: Boundary
    """The boundary of an endpoint.

    `"("` is open and left, `")"` is open and right,`"["` is closed and left, and
    `"]"` is closed and right.
    """

    def __lt__(self, other: Endpoint) -> bool:
        if not isinstance(other, Endpoint):
            return NotImplemented

        if self.value != other.value:
            return self.value < other.value

        return _BOUNDARY_ORDER[self.boundary] < _BOUNDARY_ORDER[other.boundary]

    def __eq__(self, other: Endpoint) -> bool:
        if not isinstance(other, Endpoint):
            return NotImplemented

        return (
            self.value == other.value
            and _BOUNDARY_ORDER[self.boundary] == _BOUNDARY_ORDER[other.boundary]
        )

    def __str__(self):
        if self.is_left:
            return f"{self.boundary}{self.value}"
        return f"{self.value}{self.boundary}"

    @property
    def is_closed(self) -> bool:
        return self.boundary in "[]"

    @property
    def is_open(self) -> bool:
        return self.boundary in "()"

    @property
    def is_left(self) -> bool:
        return self.boundary in "[("

    @property
    def is_right(self) -> bool:
        return self.boundary in ")]"


def _merge(
    a: list[Endpoint], b: list[Endpoint], op: Callable[[bool, bool], bool]
) -> list[Endpoint]:
    """Merge two sorted lists of endpoints with a given set operation.

    This is a sweep-line algorithm; as each endpoint is encountered one of
    `inside_a` or `inside_b` is flipped depending on whether the point belongs
    to `a` or `b`. This may flip `inside_region` (depending on `op`) which adds
    a new endpoint to the output.
    """
    endpoints: list[Endpoint] = []
    i: int = 0
    j: int = 0
    inside_a: bool = False
    inside_b: bool = False
    inside_region: bool = False

    while i < len(a) or j < len(b):
        if i >= len(a):
            current_min = current_b = b[j]
            current_a = None
        elif j >= len(b):
            current_min = current_a = a[i]
            current_b = None
        else:
            current_a = a[i]
            current_b = b[j]
            current_min = min(current_a, current_b)

        if current_a == current_min:
            inside_a = not inside_a
            i += 1

        if current_b == current_min:
            inside_b = not inside_b
            j += 1

        if op(inside_a, inside_b) != inside_region:
            inside_region = not inside_region

            # Boundary types can swap when differencing depending on
            # whether the endpoint is inside a region.
            is_closed = current_min.is_closed
            b_in_a = inside_a and current_b == current_min
            a_in_b = inside_b and current_a == current_min
            if op is sub and b_in_a or op is xor and (a_in_b or b_in_a):
                is_closed = not is_closed

            boundary = (")(", "][")[is_closed][len(endpoints) % 2 == 0]

            if (
                len(endpoints) > 0
                and endpoints[-1].value == current_min.value
                and endpoints[-1].boundary + boundary not in {"[]", ")("}
            ):  # Remove redundant endpoints such as `0), [0` or `0], [0`.
                endpoints.pop()
            else:
                endpoints.append(Endpoint(current_min.value, boundary))

    return endpoints


@dataclass
class Gaps[T: SupportsLessThan]:
    """A set of mutually exclusive continuous intervals.

    `Gaps` are created with an ordered sequence of endpoints or values with alternating
    endpoints representing the left or right endpoint of an interval. Endpoints that are
    just values and not of type `Endpoint` will be converted to an `Endpoint` with a closed
    boundary so that ::

        Gaps([1, 2]) == Gaps([Endpoint(1, "["), Endpoint(2, "]")])

    is true.
    """

    endpoints: list[T | Endpoint[T]] = field(default_factory=list)

    def __post_init__(self) -> None:
        if len(self.endpoints) % 2 == 1:
            raise ValueError("Need an even number of endpoints.")

        for i, endpoint in enumerate(self.endpoints):
            if not isinstance(endpoint, Endpoint):
                self.endpoints[i] = Endpoint(endpoint, "[" if i % 2 == 0 else "]")
            elif i % 2 == 0 and endpoint.is_right:
                raise ValueError(f"Expected left boundary, got {endpoint!r}.")
            elif i % 2 == 1 and endpoint.is_left:
                raise ValueError(f"Expected right boundary, got {endpoint!r}.")

        i = 0
        while i < len(self.endpoints) - 1:
            a = self.endpoints[i]
            b = self.endpoints[i + 1]
            if a > b:
                raise ValueError(f"Endpoints unsorted. {a!r} > {b!r}")
            if a.value == b.value and a.boundary + b.boundary not in {"[]", ")("}:
                # Intervals aren't minimally expressed, but can be fixed by removing
                # these endpoints.
                del self.endpoints[i : i + 2]
            else:
                i += 1

    @classmethod
    def from_string(cls, gaps: str) -> Self:
        """Create gaps from a string.

        Values can only be int or float. Uses standard interval notation, i.e., `"{(-inf, 1], [2, 3)}"`.
        """
        if gaps[0] != "{" or gaps[-1] != "}":
            raise ValueError(
                "Gap string must start and end with curly braces ('{', '}')."
            )

        endpoints = gaps[1:-1].replace(" ", "").split(",")
        if len(endpoints) == 1:
            return cls([])

        for i, endpoint in enumerate(endpoints):
            if endpoint.startswith(("(", "[")):
                boundary = endpoint[0]
                value = endpoint[1:]
            elif endpoint.endswith((")", "]")):
                boundary = endpoint[-1]
                value = endpoint[:-1]
            else:
                raise ValueError(f"Invalid endpoint ({endpoint!r}).")

            try:
                value = int(value)
            except ValueError:
                value = float(value)

            endpoints[i] = Endpoint(value, boundary)

        return cls(endpoints)

    def __or__(self, other: Gaps) -> Self:
        if not isinstance(other, Gaps):
            return NotImplemented

        return type(self)(_merge(self.endpoints, other.endpoints, or_))

    def __and__(self, other: Gaps) -> Self:
        if not isinstance(other, Gaps):
            return NotImplemented

        return type(self)(_merge(self.endpoints, other.endpoints, and_))

    def __xor__(self, other: Gaps) -> Self:
        if not isinstance(other, Gaps):
            return NotImplemented

        return type(self)(_merge(self.endpoints, other.endpoints, xor))

    def __sub__(self, other: Gaps) -> Self:
        if not isinstance(other, Gaps):
            return NotImplemented

        return type(self)(_merge(self.endpoints, other.endpoints, sub))

    def __bool__(self) -> bool:
        return len(self.endpoints) > 0

    def __contains__(self, value: T) -> bool:
        i = bisect(self.endpoints, value, key=attrgetter("value"))
        if i == 0:
            return False

        if self.endpoints[i - 1].value == value:
            return self.endpoints[i - 1].is_closed

        if i < len(self.endpoints):
            return self.endpoints[i].is_right

        return False

    def __str__(self) -> str:
        return f"{{{", ".join(str(endpoint) for endpoint in self.endpoints)}}}"



import pytest

from mind_the_gaps import Endpoint, Gaps


def test_singleton_ok():
    Gaps([1, 1])


def test_missing_singleton_ok():
    Gaps([0, Endpoint(1, ")"), Endpoint(1, "("), 2])


def test_fix_not_minimally_expressed():
    assert Gaps([0, Endpoint(1, ")"), 1, 2]) == Gaps([0, 2])
    assert Gaps([0, 1, Endpoint(1, "("), 2]) == Gaps([0, 2])
    assert Gaps([0, 0, 0, 0]) == Gaps([0, 0])
    assert Gaps([1, 2, 2, 3]) == Gaps([1, 3])


def test_unsorted_closed():
    with pytest.raises(ValueError, match="unsorted"):
        Gaps([1, 0])


def test_unsorted_open():
    with pytest.raises(ValueError, match="unsorted"):
        Gaps([Endpoint(1, "("), Endpoint(0, ")")])


def test_unsorted_open_closed():
    with pytest.raises(ValueError, match="unsorted"):
        Gaps([Endpoint(0, "("), 0, 1, 2])


def test_unsorted_closed_open():
    with pytest.raises(ValueError, match="unsorted"):
        Gaps([0, Endpoint(0, ")"), 1, 2])


def test_unsorted_open_open():
    with pytest.raises(ValueError, match="unsorted"):
        Gaps([Endpoint(1, "("), Endpoint(1, ")")])


def test_wrong_boundary():
    with pytest.raises(ValueError, match="left"):
        Gaps([Endpoint(0, "]"), 1])

    with pytest.raises(ValueError, match="left"):
        Gaps([Endpoint(0, ")"), 1])

    with pytest.raises(ValueError, match="right"):
        Gaps([0, Endpoint(1, "[")])

    with pytest.raises(ValueError, match="right"):
        Gaps([0, Endpoint(1, "(")])


def test_wrong_number_of_endpoints():
    with pytest.raises(ValueError, match="endpoints"):
        Gaps([0, 1, 2])


def test_gap_bool():
    assert bool(Gaps([0, 1]))
    assert not bool(Gaps())


def test_gap_str():
    assert str(Gaps([0, 1])) == "{[0, 1]}"
    assert str(Gaps([Endpoint(0, "("), 1])) == "{(0, 1]}"
    assert str(Gaps([0, Endpoint(1, ")")])) == "{[0, 1)}"
    assert str(Gaps([Endpoint(0, "("), Endpoint(1, ")")])) == "{(0, 1)}"


def test_gap_not_implemented():
    with pytest.raises(TypeError):
        Gaps() | None

    with pytest.raises(TypeError):
        Gaps() - None

    with pytest.raises(TypeError):
        Gaps() & None

    with pytest.raises(TypeError):
        Gaps() ^ None



import pytest

from mind_the_gaps import Endpoint, Gaps


def test_malformed_string():
    with pytest.raises(ValueError, match="string must start and end"):
        Gaps.from_string("[0, 1]")

    with pytest.raises(ValueError, match="Invalid endpoint"):
        Gaps.from_string("{[0, 1, 2}")


def test_from_string_closed():
    assert Gaps.from_string("{[0, 1], [2, 3]}") == Gaps([0, 1, 2, 3])


def test_from_string_open():
    assert Gaps.from_string("{(0, 1), (2, 3)}") == Gaps(
        [Endpoint(0, "("), Endpoint(1, ")"), Endpoint(2, "("), Endpoint(3, ")")]
    )


def test_from_string_half_open():
    assert Gaps.from_string("{(-inf, 0], [1, 2)}") == Gaps(
        [Endpoint(-float("inf"), "("), 0, 1, Endpoint(2, ")")]
    )


def test_from_string_singleton():
    assert Gaps.from_string("{[0, 0]}") == Gaps([0, 0])


def test_bounded_missing_singleton():
    assert Gaps.from_string("{[0, 1), (1, 2]}") == Gaps(
        [0, Endpoint(1, ")"), Endpoint(1, "("), 2]
    )


def test_from_string_empty():
    assert Gaps.from_string("{}") == Gaps()



from mind_the_gaps import Endpoint, Gaps, x


def test_x_str():
    assert str(x) == "x"


def test_x_closed():
    assert Gaps([0 <= x, x <= 1, 2 <= x, x <= 3]) == Gaps([0, 1, 2, 3])


def test_x_open():
    assert Gaps([0 < x, x < 1, 2 < x, x < 3]) == Gaps(
        [Endpoint(0, "("), Endpoint(1, ")"), Endpoint(2, "("), Endpoint(3, ")")]
    )


def test_x_singleton():
    assert Gaps([0 <= x, x <= 0]) == Gaps([0, 0])


def test_missing_singleton():
    assert Gaps([0 <= x, x < 1, 1 < x, x <= 2]) == Gaps(
        [0, Endpoint(1, ")"), Endpoint(1, "("), 2]
    )



from mind_the_gaps import Endpoint, Gaps


def test_bounded_disjoint():
    a = Gaps([0, 1])
    b = Gaps([2, 3])
    assert a & b == Gaps()


def test_unbounded_disjoint():
    a = Gaps([-float("inf"), 0])
    b = Gaps([1, 2])
    assert a & b == Gaps()


def test_bounded_intersecting_proper():
    a = Gaps([0, 2])
    b = Gaps([1, 3])
    assert a & b == Gaps([1, 2])


def test_bounded_intersecting_singleton():
    a = Gaps([0, 2])
    b = Gaps([1, 1])
    assert a & b == b


def test_bounded_missing_singleton():
    a = Gaps([0, 2])
    b = Gaps([0, Endpoint(1, ")"), Endpoint(1, "("), 2])
    assert a & b == b


def test_empty_intersection():
    a = Gaps([0, 2])
    b = Gaps()
    assert a & b == b



from mind_the_gaps import Endpoint, Gaps


def test_left_open_not_contained():
    assert 0 not in Gaps([Endpoint(0, "("), 1])


def test_left_closed_contained():
    assert 0 in Gaps([0, 1])


def test_right_open_not_contained():
    assert 1 not in Gaps([0, Endpoint(1, ")")])


def test_right_closed_contained():
    assert 1 in Gaps([0, 1])


def test_value_less_than_first_endpoint_not_containted():
    assert -1 not in Gaps([0, 1])


def test_value_greater_than_last_endpoint_not_contained():
    assert 7 not in Gaps([0, 1])


def test_value_not_equal_to_endpoint_but_contained():
    assert 0.5 in Gaps([0, 1])


def test_value_not_equal_to_endpoint_but_not_contained():
    assert 1.5 not in Gaps([0, 1, 2, 3])



import pytest

from mind_the_gaps import Endpoint


def test_endpoint_str():
    assert str(Endpoint(0, ")")) == "0)"
    assert str(Endpoint(0, "]")) == "0]"
    assert str(Endpoint(0, "(")) == "(0"
    assert str(Endpoint(0, "[")) == "[0"


def test_endpoint_properties():
    assert Endpoint(0, ")").is_open
    assert Endpoint(0, ")").is_right
    assert Endpoint(0, "(").is_open
    assert Endpoint(0, "(").is_left
    assert Endpoint(0, "[").is_closed
    assert Endpoint(0, "[").is_left
    assert Endpoint(0, "]").is_closed
    assert Endpoint(0, "]").is_right


def test_endpoint_lt_not_implemented():
    with pytest.raises(TypeError):
        Endpoint(0, "(") < 0



"""A library for unions, intersections, subtractions, and xors of intervals (gaps)."""

from .gaps import Endpoint, Gaps
from .var import Var, x

__all__ = ["Endpoint", "Gaps", "Var", "x"]

__version__ = "0.3.4"
